/*
 * C6Run_control.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Defines DSP/Bios main entry point that serves to kick off the        *
*  user supplied main in its own task.                                  *
************************************************************************/


/*  ----------------------------------- Headers for User Code       */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>

/*  ----------------------------------- DSP/BIOS Headers            */
#include "C6Runcfg.h"
#include <sys.h>
#include <sem.h>
#include <msgq.h>
#include <pool.h>
#include <gbl.h>
#include <clk.h>
#include <log.h>
#include <tsk.h>

/*  ----------------------------------- DSP/BIOS LINK Headers       */
#include <dsplink.h>
#include <failure.h>

/*  ----------------------------------- Headers for CIO Code        */
#include "control_ipc.h"
#include "C6Run_rpc.h"

/*  ----------------------------------- This module's header        */
#include "C6Run_control.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/

// Declaration for main function from either user-supplied application 
// code (for C6RunApp) or from a dummy implementation (for C6RunLib)
extern int main(int argc, char* argv []);


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

/* FILEID is used by SET_FAILURE_REASON macro. */
#define FILEID  FID_APP_C


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/


/************************************************************
* Local Variable Definitions                                *
************************************************************/

// We use static global variable for the message pointer
static CONTROL_Msg *msg;

static bool LOCAL_isModuleInitialized = false;


/************************************************************
* Global Variable Definitions                               *
************************************************************/


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Task function for DSP-side of C6Run control
void C6RUN_CONTROL_task(int argc, char* argv [])
{
  uint16_t msgId;
  int retVal = 0;

  // Setup inter-processor interfaces
  int32_t status = C6RUN_CONTROL_init();
  
  if (status != SYS_OK)
  {
    exit(1);
  }
  
  // Wait for the STARTLIB or STARTAPP message
  msgId = CONTROL_IPC_recvMsg(&msg);
  
  if (CONTROL_DSP_STARTLIB == msgId)
  {
    // Task handler for TSK_create
    TSK_Handle task;
    TSK_Attrs  attrs = TSK_ATTRS;  
  
    // Start RPC task
    task = TSK_create((Fxn)C6RUN_RPC_task, &attrs);
    if (task == NULL)
    {
      SET_FAILURE_REASON(SYS_EALLOC);
      LOG_error("Create C6RUN_RPC_task: Failed", 0);
    }
  
    // Send the DSP_READY message so the GPP knows we are waiting
    CONTROL_IPC_sendMsg(msg, CONTROL_DSP_READY);
    
    // Wait for the stop message (and ONLY the stop message)
    while (1)
    {
      msgId = CONTROL_IPC_recvMsg(&msg);
      if (CONTROL_DSP_STOP == msgId)
      {
        break;
      }
    }
    
    // Delete the RPC task
    TSK_delete(task);
  }
  else if (CONTROL_DSP_STARTAPP == msgId)
  {
    // Call the user supplied main function
    retVal = main(argc, argv);
  }
  
  // We call the exit routine as this should be the same as what happens
  // when main() terminates.  It ensures that all atexit() registered routines
  // will get called.
  exit(retVal);
}

// Init the CONTROL routines
int32_t C6RUN_CONTROL_init( void )
{
  if (!LOCAL_isModuleInitialized)
  {
    int32_t status = SYS_OK;
    status |= CONTROL_IPC_init();
    status |= C6RUN_CIO_init();

    if (status != SYS_OK)
    {
      LOG_error("C6RUN_CONTROL_init() failed.", 0);
      return status;
    }
    
    LOCAL_isModuleInitialized = true;
  }
  return SYS_OK;
}

// Exit routine to tell the GPP host that we are exiting cleanly and 
// return a result
void C6RUN_CONTROL_exit(int retVal)
{
  // Send FINISHED message and teardown IPC and return retVal back to GPP
  memcpy ((void *)msg->buffer,(void *)&retVal,sizeof(int));
  
  // CIO exit
  C6RUN_CIO_exit();
 
  // Control message to tell the GPP to terminate the DSP 
  CONTROL_IPC_sendMsg(msg,CONTROL_DSP_FINISHED);
  
  // CONTROL_IPC_exit();
  
  // Die here waiting for the DSP to be shutdown
  while(1);
}

// Abort routine that would get called in unusual circumstances
void C6RUN_CONTROL_abort( void )
{
  // CIO exit
  C6RUN_CIO_exit();

  CONTROL_IPC_sendMsg(msg,CONTROL_DSP_ABORTED);
  
  // CONTROL_IPC_exit();

  // Die here waiting for the DSP to be shutdown  
  while(1);  
}


/************************************************************
* Local Function Definitions                                *
************************************************************/


/***********************************************************
* End file                                                 *
***********************************************************/
